/***************************************************************************
 *   Copyright (C) 2008 by Christoph Burger-Scheidlin                      *
 *   mail@christoph.burger-scheidlin.name                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "countdown.h"
#include <QPainter>

#include <KGlobalSettings>
#include <KColorScheme>
#include <KLocale>
#include <KSystemTimeZone>

#include <plasma/theme.h>

const int secondsInYear  = 31536000;
const int secondsInMonth =  2592000;
const int secondsInWeek  =   604800;
const int secondsInDay   =    86400;
const int secondsInHour  =     3600;
const int secondsInMin   =       60;

const int monthLine  = secondsInYear * 3 / 2;
const int weekLine   = secondsInMonth * 3 / 2;
const int dayLine    = secondsInWeek * 3 / 2;
const int hourLine   = secondsInDay * 3 / 2;
const int minuteLine = secondsInHour * 3 / 2;
const int secondLine = secondsInMin * 3 / 2;

Countdown::Countdown(QObject *parent, const QVariantList &args)
	: Plasma::Applet(parent, args),
	  m_font(KGlobalSettings::generalFont()),
	  m_color(),
	  m_negativeColor(),
	  m_fontBold(false),
	  m_fontItalic(false),
	  m_negative(true),
	  m_targetDateTime(),
	  m_currentAlignment(Plasma::NoAlignment),
	  m_configDialog(0)
{
	setDrawStandardBackground(true);
	setHasConfigurationInterface(true);
}

Countdown::~Countdown()
{
	delete m_configDialog;
}

void Countdown::init()
{
	KConfigGroup cg = config();

	m_targetDateTime = cg.readEntry("targetDateTime", QDateTime());
	m_timezone = cg.readEntry("timezone", "Local");

	//FIXME: currently (4.0.0) foreground color seems to be black
//	m_color = KColorScheme(QPalette::Active, KColorScheme::View, Plasma::Theme::self()->colors()).foreground().color();
	m_color = Qt::white;
	m_color = cg.readEntry("color", m_color);
	m_negativeColor = Qt::red;
	m_negativeColor = cg.readEntry("negativeColor", m_negativeColor);

	m_font = cg.readEntry("font", m_font);
	m_fontBold = cg.readEntry("fontBold", true);
	m_fontItalic = cg.readEntry("fontItalic", false);
	m_font.setBold(m_fontBold);
	m_font.setItalic(m_fontItalic);

	if ( m_targetDateTime.isNull() ) {
		setNeedsConfiguring( true );
	}

	setMinimumContentSize(40, 50);

	m_currentAlignment = Plasma::NoAlignment;
	dataEngine("time")->connectSource("UTC", this, 1000, Plasma::NoAlignment);
}

void Countdown::paintInterface(QPainter *p,
	const QStyleOptionGraphicsItem *option, const QRect &contentsRect)
{
	Q_UNUSED( option );

// 	p->save();

	if ( m_negative ) {
		p->setPen( QPen( m_negativeColor ) );
	} else {
		p->setPen(QPen(m_color));
	}
	p->setRenderHint(QPainter::SmoothPixmapTransform);
	p->setRenderHint(QPainter::Antialiasing);

	m_font.setBold(m_fontBold);
	m_font.setItalic(m_fontItalic);
	m_font.setPointSize(qMax((int)(contentsRect.height()/1.5), 1));

	QRect displayRect( contentsRect );
	if ( m_negative ) {
		QRect negativeRect( 0, contentsRect.top(), contentsRect.right(), contentsRect.height() / 4 );
		QString negativeString = i18nc( "Indicate that the countdown has crossed "
						"over its line, any time displayed is past "
						"the target time",
      						"overdue");
		QRect overdueRect = preparePainter(p, negativeRect, m_font, negativeString);
		displayRect.setTop( contentsRect.top()+overdueRect.height() );
		p->drawText(QRectF(0,
			contentsRect.top(),
			contentsRect.right(),
			overdueRect.height()),
			negativeString,
			QTextOption(Qt::AlignHCenter)
		);

	}

	preparePainter( p, displayRect, m_font, m_displayString );

	p->drawText(displayRect,
		m_displayString,
		QTextOption(Qt::AlignCenter)
		);
// 	p->restore();
}

void Countdown::constraintsUpdated(Plasma::Constraints)
{
	// I'm not sure why this is needed, since we don't actually change
	// any geometry in this method, but it screws up the system tray
	// without it.
	updateGeometry();
}

void Countdown::dataUpdated(const QString &source, const Plasma::DataEngine::Data &data)
{
	Q_UNUSED(source);
	QDateTime utc(data["Date"].toDate(), data["Time"].toTime(), Qt::UTC);

	int secs = utc.secsTo( m_targetDateTime );
	m_displayString = determineDisplayString( secs );

	if ( secs > 0 and secs <= secondLine + secondsInMin and m_currentAlignment == Plasma::AlignToMinute ) {
		m_currentAlignment = Plasma::NoAlignment;
		dataEngine("time")->connectSource("UTC", this, 1000, Plasma::NoAlignment);
	} else if (  m_currentAlignment == Plasma::NoAlignment
		 and ( secs < -1*secondLine or secs > secondLine + secondsInMin )
		  ) {
		m_currentAlignment = Plasma::AlignToMinute;
		dataEngine("time")->connectSource("UTC", this, 60000, Plasma::AlignToMinute);
	}

	if (m_displayString != m_lastDisplayString) {
		m_lastDisplayString = m_displayString;
		update();
	}
}

QString Countdown::determineDisplayString( int seconds )
{
	m_negative = seconds < 0;
	seconds = qAbs(seconds);

	QString display;
	if ( seconds > monthLine ) {
		display = i18nc("@label display years: "
				"%1 number of years",
				"%1 years", ( seconds + secondsInYear - 1 )/ secondsInYear);
	} else if ( seconds > weekLine ) {
		display = i18nc("@label display months: "
				"%1 number of months",
				"%1 months", ( seconds + secondsInMonth - 1 )/ secondsInMonth);
	} else if ( seconds > dayLine ) {
		display = i18nc("@label display weeks: "
				"%1 number of weeks",
				"%1 weeks", ( seconds + secondsInWeek - 1 )/ secondsInWeek);
	} else if ( seconds > hourLine ) {
		display = i18nc("@label display days: "
				"%1 number of days",
				"%1 days", ( seconds + secondsInDay - 1 )/ secondsInDay);
	} else if ( seconds > minuteLine ) {
		display = i18nc("@label display hours: "
				"%1 number of hours",
				"%1 hours", ( seconds + secondsInHour - 1)/ secondsInHour);
	} else if ( seconds > secondLine ) {
		display = i18nc("@label display minutes: "
				"%1 number of minutes",
				"%1 min", ( seconds + secondsInMin - 1 )/ secondsInMin);
	} else {
		display = i18nc("@label display seconds: "
				"%1 number of seconds",
				"%1 sec", seconds);
	}

	return display;
}

void Countdown::showConfigurationInterface()
{
	if (m_configDialog == 0) {
		m_configDialog = new KDialog;

		m_configDialog->setCaption(i18n("Configure Countdown"));

		QWidget *widget = new QWidget;
		ui.setupUi(widget);
		m_configDialog->setMainWidget(widget);
		m_configDialog->setButtons( KDialog::Ok | KDialog::Cancel | KDialog::Apply );

		connect( m_configDialog, SIGNAL(applyClicked()), this, SLOT(configAccepted()) );
		connect( m_configDialog, SIGNAL(okClicked()), this, SLOT(configAccepted()) );
	}

	ui.fontBold->setChecked( m_fontBold );
	ui.fontItalic->setChecked( m_fontItalic );
	ui.font->setCurrentFont( m_font );
	ui.color->setColor( m_color );
	ui.negativeColor->setColor( m_negativeColor );

	if ( m_targetDateTime.isNull() ) {
		QDateTime now = QDateTime::currentDateTime();
		ui.targetDateTime->setDateTime(QDateTime(QDate(now.date()), QTime(now.time().hour(), now.time().minute(), 0)));
	} else {
		ui.targetDateTime->setDateTime( m_targetDateTime );
	}

	ui.timeZones->setSelected(m_timezone, true);
	ui.timeZones->setEnabled(m_timezone != "Local");
	ui.differentTimeZone->setChecked(m_timezone != "Local");

	m_configDialog->show();
}

void Countdown::configAccepted()
{
	KConfigGroup cg = config();

	m_font = ui.font->currentFont();
	m_color = ui.color->color();
	m_negativeColor = ui.negativeColor->color();
	m_fontBold = ui.fontBold->checkState() == Qt::Checked;
	m_fontItalic = ui.fontItalic->checkState() == Qt::Checked;
	QDateTime targetDateTime = ui.targetDateTime->dateTime();

	QStringList tzs = ui.timeZones->selection();
	if (ui.differentTimeZone->checkState() == Qt::Checked and tzs.count() > 0) {
		m_timezone = tzs.at(0);
	} else {
		m_timezone = "Local";
	}

	m_targetDateTime = targetTimeInUTC( targetDateTime, m_timezone );

	m_font.setBold(m_fontBold);
	m_font.setItalic(m_fontItalic);

	cg.writeEntry("font", m_font);
	cg.writeEntry("color", m_color);
	cg.writeEntry("negativeColor", m_negativeColor);
	cg.writeEntry("fontBold", m_fontBold);
	cg.writeEntry("fontItalic", m_fontItalic);
	cg.writeEntry("targetDateTime", m_targetDateTime);
	cg.writeEntry("timezone", m_timezone);

	setNeedsConfiguring( false );

	m_currentAlignment = Plasma::NoAlignment;
	dataEngine("time")->connectSource("UTC", this, 1000, Plasma::NoAlignment);
	emit configNeedsSaving();
	update();
}

QDateTime Countdown::targetTimeInUTC( const QDateTime& dateTime, const QString& timezone )
{
	QDateTime output;
	if ( timezone == "Local" ){
		output = KSystemTimeZones::local().toUtc( dateTime );
	} else {
		output = KSystemTimeZones::zone( timezone ).toUtc(dateTime);
	}
	return output;
}

QRect Countdown::preparePainter(QPainter *p, const QRect &rect, const QFont &font, const QString &text)
{
    QRect tmpRect;
    QFont tmpFont = font;

    // Starting with the given font, decrease its size until it'll fit in the
    // given rect allowing wrapping where possible
    do {
        p->setFont(tmpFont);
        tmpFont.setPointSize(tmpFont.pointSize() - 1);
        tmpRect = p->boundingRect(rect, Qt::TextWordWrap, text);
    } while (tmpFont.pointSize() >= 1 && (tmpRect.width() > rect.width() ||
            tmpRect.height() > rect.height()));

    return tmpRect;
}

#include "countdown.moc"
